{-# OPTIONS --cubical #-}

 -- to allow opening this module in other files while there are still holes
{-# OPTIONS --allow-unsolved-metas #-}


module Common.Common where

open import Cubical.Foundations.Prelude
open import Cubical.Foundations.Function
open import Cubical.Foundations.HLevels
open import Cubical.Data.Nat hiding (_^_) renaming (ℕ to Nat)
open import Cubical.Data.Sum
open import Cubical.Data.Empty
open import Cubical.Data.Sigma
open import Cubical.Data.Unit renaming (Unit to ⊤)

open import Cubical.Algebra.Monoid.Base
open import Cubical.Algebra.Monoid.FreeCommProduct
open import Cubical.Algebra.CommMonoid.Base


private
  variable
    ℓ ℓ' ℓ'' : Level
    ℓX ℓY ℓZ : Level
    ℓA ℓB ℓC ℓD : Level
    ℓR ℓS : Level

id : {ℓ : Level} -> {A : Type ℓ} -> A -> A
id x = x

_^_ : {ℓ : Level} -> {A : Type ℓ} -> (A -> A) -> Nat -> A -> A
f ^ zero = id
f ^ suc n = f ∘ (f ^ n)

⊥*→⊥ : {ℓ : Level} -> ⊥* {ℓ} -> ⊥
⊥*→⊥ ()


inl≠inr : ∀ {ℓ ℓ' : Level} {A : Type ℓ} {B : Type ℓ'} ->
  (a : A) (b : B) -> inl a ≡ inr b -> ⊥
inl≠inr {_} {_} {A} {B} a b eq = transport (cong (diagonal ⊤ ⊥) eq) tt
  where
    diagonal : (Left Right : Type) -> (A ⊎ B) -> Type
    diagonal Left Right (inl a) = Left
    diagonal Left Right (inr b) = Right



-- Definitions about relations and two-cells

compRel : {ℓ' : Level} ->
  {X : Type ℓX} {Y : Type ℓY} {Z : Type ℓZ}
  (R1 : X -> Y -> Type ℓ') ->
  (R2 : Y -> Z -> Type ℓ') ->
  (X -> Z -> Type (ℓ-max ℓY ℓ'))
compRel {ℓ' = ℓ'} {X = X} {Y = Y} {Z = Z} R1 R2 x z = Σ[ y ∈ Y ] R1 x y × R2 y z

isPropValuedRel : {A : Type ℓA} {B : Type ℓB} ->
  (R : A -> B -> Type ℓR) -> Type (ℓ-max (ℓ-max ℓA ℓB) ℓR)
isPropValuedRel {A = A} {B = B} R =
  (x : A) -> (y : B) -> isProp (R x y)

TwoCell : {A : Type ℓA} {B : Type ℓB} {C : Type ℓC} {D : Type ℓD} ->
-- {A C : Type ℓ} -> {B D : Type ℓ'}
  (R : A -> B -> Type ℓR) ->
  (S : C -> D -> Type ℓS)
  (f : A -> C) ->
  (g : B -> D) ->
  Type (ℓ-max (ℓ-max (ℓ-max ℓA ℓB) ℓR) ℓS)
-- Type (ℓ-max (ℓ-max (ℓ-max ℓ ℓ') ℓR) ℓS)
TwoCell R S f g = ∀ a b -> R a b -> S (f a) (g b)

isPropTwoCell : {A : Type ℓA} {B : Type ℓB} {C : Type ℓC} {D : Type ℓD} ->
  {R : A -> B -> Type ℓR} ->
  {S : C -> D -> Type ℓS}
  {f : A -> C} ->
  {g : B -> D} ->
  isPropValuedRel S ->
  isProp (TwoCell R S f g)
isPropTwoCell isPropValuedS =
  isPropΠ3 (λ a b aRb -> isPropValuedS _ _)





isSplitMono : {ℓA ℓB : Level} {A : Type ℓA} {B : Type ℓB} ->
  (f : A -> B) -> Type (ℓ-max ℓA ℓB)
isSplitMono {A = A} {B = B} f =
 Σ[ g ∈ (B -> A) ] (∀ a -> g (f a) ≡ a)


InjectiveMonoidHom : {ℓm ℓn : Level} ->
  (M : Monoid ℓm) (N : Monoid ℓn) -> Type (ℓ-max ℓm ℓn)
InjectiveMonoidHom M N =
  Σ[ f ∈ MonoidHom M N ] isSplitMono (fst f)

